/*
Copyright (C) 2012  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
(version 2) as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*
  TODO
  * Mabye use map structure to track watchlist descriptors and filepaths.
  * Document (Doxygen?)
  * Homogenize debug messages and maybe add indentation for nested calls.

  NOTES
  According the the man page, files can be created in new subdirectories before
  the subdirectory creation event is received/processed. The directory tree
  must therefore be rescanned when directories are added or moved.
*/

#include <dirent.h>
#include <errno.h>
#include <grp.h>
#include <limits.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/inotify.h>
#include <sys/stat.h>
#include <linux/sysctl.h>
#include <unistd.h>

#include "version.h"

#ifndef DEBUG
  #define DEBUG 0
#endif

#define debug_print(fmt, ...) \
do \
{ \
  if (DEBUG) \
    fprintf(\
      stderr, \
      "%s:%d:%s(): " fmt, \
      __FILE__, \
      __LINE__, \
      __func__, \
      __VA_ARGS__ \
    ); \
} while (0)

#define STR_EXPAND(tok) #tok
#define STR(tok) STR_EXPAND(tok)

#define NAME "maown"

#define EVENT_SIZE  (sizeof(struct inotify_event))
#define EVENT_Q_LEN (16384 * ( EVENT_SIZE + 16 ))
#define EVENTS (IN_CREATE | IN_ATTRIB | IN_DELETE_SELF | IN_MOVED_TO | IN_MOVE_SELF | IN_Q_OVERFLOW | IN_DONT_FOLLOW | IN_ONLYDIR)

#define DELTA_WL 100
#define EMPTY_SPACE_THRESHOLD 500




// watched_dir
typedef struct
{
  int wd;
  char * path;
} wdir_t;



// global watchlist for tracking filepaths
typedef struct
{
  int fd;
  int events;
  int size;
  int used;
  wdir_t * list;
} watchlist_t;





/*
  Read int from file.
*/
int
read_int(char * path)
{
  debug_print("reading int from %s\n", path);
  FILE * f;
  int i;
  f = fopen(path, "rb");
  if (f == NULL)
  {
    fprintf(stderr, "error: failed to open %s\n", path);
    exit(EXIT_FAILURE);
  }
  fscanf(f, "%d", &i);
  fclose(f);
  return i;
}



/*
  Return index of given watch descriptor.

  TODO
  Improve efficiency.
*/
int
get_wd_index(int wd, watchlist_t * watchlist)
{
  debug_print("%d, %p\n", wd, watchlist);
  int i;
  debug_print(
    "watchlist.used: %d, watchlist.size: %d\n",
    watchlist->used,
    watchlist->size
  );
  for(i = 0; i < watchlist->used; i++)
  {
    if (watchlist->list[i].wd == wd)
    {
      debug_print("returned %d\n", i);
      return i;
    }
  }
  debug_print("returned %d\n", -1);
  return -1;
}



/*
  Return the path associated with a watch descriptor.
*/
char *
get_wd_path(int wd, watchlist_t * watchlist)
{
  debug_print("%d, %p\n", wd, watchlist);
  int i;
  i = get_wd_index(wd, watchlist);
  return watchlist->list[i].path;
}



/*
  Remove unused space from the watchlist by reallocating the buffer. This
  should only be called when there is a significant fraction of unused space.

  Some space will be left to accommodate further items.
*/
void
maybe_free_unused_space(watchlist_t * watchlist)
{
  int newsize, bsize;
  void * newptr;

  if (watchlist->size - watchlist->used > EMPTY_SPACE_THRESHOLD)
  {
    newsize = (watchlist->used + DELTA_WL);
    bsize = newsize * sizeof(wdir_t);
    debug_print("shrinking watchlist (%d -> %d)\n", watchlist->size, newsize);
    newptr = realloc(watchlist->list, bsize);
    if (newptr == NULL)
    {
      perror("maybe_free_unused_space");
      exit(EXIT_FAILURE);
    }
    watchlist->list = newptr;
    watchlist->size = newsize;
  }
}



/*
  Remove item from list. This is used for IN_DELETE_SELF events as the
  deleted directory is automatically unwatched.

  TODO
  See list for get_wd_index.
*/
void
remove_wd(int wd, watchlist_t * watchlist)
{
  debug_print(
    "wd: %d, watchlist %p (%d/%d)\n",
    wd, watchlist, watchlist->used, watchlist->size
  );
  int i;
  for(i = 0; i < watchlist->used; i++)
  {
    if (watchlist->list[i].wd == wd)
    {
      free(watchlist->list[i].path);
      // Move the last element in the list to the new empty slot.
      watchlist->list[i] = watchlist->list[-- watchlist->used];
      maybe_free_unused_space(watchlist);
      return;
    }
  }
}



/*
  Add directory to watchlist.
*/
void
watch_dir(char * path, watchlist_t * watchlist)
{
  debug_print("%s, %p\n", path, watchlist);
  void * newptr;
  int bsize, bdelta, wd, i;
  char * pathcpy;

  wd = inotify_add_watch(watchlist->fd, path, watchlist->events);
  i = get_wd_index(wd, watchlist);

  // already in list, nothing to do
  if (i > -1)
  {
    return;
  }

  if (watchlist->used == watchlist->size)
  {
    debug_print("expanding watchlist (%d += " STR(DELTA_WL) ")\n", watchlist->size);
    bsize = watchlist->size * sizeof(wdir_t);
    bdelta = DELTA_WL * sizeof(wdir_t);
    // catch overflows
    if (INT_MAX - bsize < bdelta)
    {
      fprintf(stderr, "error: watchlist out of range");
      exit(EXIT_FAILURE);
    }
    newptr = realloc(watchlist->list, bsize + bdelta);
    if (newptr == NULL)
    {
      perror("watch_dir");
      exit(EXIT_FAILURE);
    }
    watchlist->list = newptr;
    watchlist->size += DELTA_WL;
    debug_print(
      "watchlist.used: %d, watchlist.size: %d\n",
      watchlist->used,
      watchlist->size
    );
  }
  pathcpy = malloc(strlen(path)+1);
  strcpy(pathcpy, path);
  wdir_t newdir = {.wd = wd, .path = pathcpy};
  watchlist->list[watchlist->used++] = newdir;
  debug_print("added %s to watchlist\n", pathcpy);
}



/*
  Remove directory tree from watchlist.
*/
void
unwatch_dir(int wd, watchlist_t * watchlist)
{
  debug_print("%d, %p\n", wd, watchlist);
  int i, l;
  char * path;

  i = get_wd_index(wd, watchlist);

  // remove this right away to avoid further events, e.g. if moved out of dir
  inotify_rm_watch(watchlist->fd, watchlist->list[i].wd);

  path = watchlist->list[i].path;
  l = strlen(path);

  // move up last element of list and reduce size
  watchlist->list[i] = watchlist->list[-- watchlist->used];

  // unwatch all subdirectories
  // TODO: improve efficiency (sorted paths?)
  for(i = 0; i < watchlist->used; i++)
  {
    if (strncmp(watchlist->list[i].path, path, l) == 0)
    {
      free(watchlist->list[i].path);
      watchlist->list[i] = watchlist->list[-- watchlist->used];
      i--;
    }
  }
  free(path);
  maybe_free_unused_space(watchlist);
}



/*
  Remove all watchlist entries.
*/
void
unwatch_all(watchlist_t * watchlist)
{
  debug_print("%p\n", watchlist);
  int i;
  // unwatch all subdirectories
  for(i = 0; i < watchlist->used; i++)
  {
    inotify_rm_watch(watchlist->fd, watchlist->list[i].wd);
    free(watchlist->list[i].path);
  }
  watchlist->used = 0;
}



/*
  Add trailing slash if missing. Return length of string.
*/
int
maybe_append_slash(char * path)
{
  debug_print("%s\n", path);
  int i = strlen(path);
  if (path[i-1] != '/')
  {
    path[i] = '/';
    path[i+1] = '\0';
    i++;
  }
  return i;
}


/*
  Chown files to given uid and gid if necessary, and ensure that group
  permissions match user permissions.
*/
int
adjust_attrib(char * path, uid_t uid, gid_t gid)
{
  debug_print("%s, %d, %d\n", path, uid, gid);
  mode_t mode;
  struct stat st;

  if (! lstat(path, &st))
  {
    /*
      Skip symlinks. If they point to files in the watched hierarchy then the
      target file will be updated as expected. If the link points to a file
      outside of the watched hierarchy, it should not be touched anyway.
    */
    if (S_ISLNK(st.st_mode))
    {
      return 0;
    }
    if (st.st_uid != uid || st.st_gid != gid)
    {
      debug_print("chowning %s\n", path);
      if (chown(path, uid, gid))
      {
        fprintf(stderr, "error: failed to chown %s (%s)\n", path, strerror(errno));
        return 1;
      }
    }
    // Replace the group mode with the user mode.
    mode = ((st.st_mode & S_IRWXU) >> 3) | (st.st_mode & ~S_IRWXG);
    if (st.st_mode != mode)
    {
      debug_print("chmoding %s\n", path);
      if (chmod(path, mode))
      {
        fprintf(stderr, "error: failed to chmod %s (%s)\n", path, strerror(errno));
        return 1;
      }
    }
    return 0;
  }
  return 1;
}


/*
  Scan directory hierarchy and chown files to uid and gid. Add all
  directories to watchlist.
*/
int
scan(char * path, watchlist_t * watchlist, uid_t uid, gid_t gid)
{
  debug_print("%s, %p, %d, %d\n", path, watchlist, uid, gid);
  char fpath[PATH_MAX+1];
  int l;
  struct dirent * de;
  struct stat st;

  if (stat(path, &st))
  {
//     fprintf(
//       stderr,
//       "scan %s, %p, %d, %d (%s)\n",
//       path, watchlist, uid, gid, strerror(errno)
//     );
//     exit(EXIT_FAILURE);
    return 0;
  }

  if (adjust_attrib(path, uid, gid))
  {
    return 1;
  }

  if (! S_ISDIR(st.st_mode))
  {
    return 0;
  }

  strcpy(fpath, path);
  l = maybe_append_slash(fpath);

  /*
    Recursion into subdirectories will probably generate uninteresting events,
    but events that we need to catch may be generated externally (e.g. file
    creation) by another app while scanning subdirectories).
  */
  watch_dir(fpath, watchlist);


  DIR * dir = opendir(path);
  if (dir == NULL)
  {
    fprintf(stderr, "failed to open directory %s\n", path);
    return 1;
  }
  while ((de = readdir(dir)) != NULL && errno == 0)
  {
    // skip "." and ".."
    if
    (
      de->d_name[0] == '.' &&
      (
        de->d_name[1] == '\0' ||
        (
          de->d_name[1] == '.' &&
          de->d_name[2] == '\0'
        )
      )
    )
    {
      continue;
    }
    strcpy(fpath+l, de->d_name);
    switch (de->d_type)
    {
      case DT_DIR:
        if (scan(fpath, watchlist, uid, gid))
        {
          return 1;
        }
        break;

      default:
        if (adjust_attrib(fpath, uid, gid))
        {
          return 1;
        }
        break;
    }
  }
  closedir(dir);
  return 0;
}



void
print_usage(FILE * fd)
{
  fprintf(
    fd,
"%s %s\n"
"usage: %s [options] <user> <group> <directory> [<directory>...]\n"
"\n"
"options:\n"
"  -d: fork process to background and print PID\n"
"  -f <path>: write PID to path\n"
"  -h: display this message and exit\n"
"\n"
    ,NAME,VERSION,NAME
  );
}



int
main(int argc, char * * argv)
{
  char * queue, * user, * group, * pid_path = NULL;
  char tmppath[PATH_MAX+1];
  int i, j, l, fd, daemonize = 0;
  struct inotify_event * e;
  struct passwd * pw;
  struct group * gr;
  uid_t uid;
  gid_t gid;
  pid_t pid;
  FILE * f;

  while((i = getopt(argc, argv, "df:h")) != -1)
  {
    switch(i)
    {
      case 'h':
        print_usage(stdout);
        return(EXIT_SUCCESS);
        break;
      case 'd':
        daemonize = 1;
        break;
      case 'f':
        pid_path = optarg;
        break;
      default:
        print_usage(stderr);
        return(EXIT_FAILURE);
        break;
    }
  }

  if (argc - optind < 3)
  {
    print_usage(stderr);
    return EXIT_FAILURE;
  }

  user = argv[optind++];
  group = argv[optind++];



  // for reading /proc/sys/fs/inotify/max_queued_events
//   int imqe_path[] = {CTL_FS, FS_INOTIFY, INOTIFY_MAX_QUEUED_EVENTS};
//   int imqe_len = sizeof(imqe_path) / sizeof(int);
  int imqe;
//   size_t imqe_size = sizeof(imqe);
//   if (sysctl(imqe_path, imqe_len, &imqe, &imqe_size, NULL, 0))
//   {
//     fprintf(
//       stderr,
//       "error: failed to determine maximum length of inotify event queue (%s)\n",
//       strerror(errno)
//     );
//     return EXIT_FAILURE;
//   }

  imqe = read_int("/proc/sys/fs/inotify/max_queued_events");


  debug_print("INOTIFY_MAX_QUEUED_EVENTS is %d\n", imqe);
  queue = malloc(imqe * ( EVENT_SIZE + 16 ));
  if (queue == NULL)
  {
    fprintf(
      stderr,
      "error: failed to allocate memory for event queue\n"
    );
    return EXIT_FAILURE;
  }

  // Get uid and gid.
  pw = getpwnam(user);
  if (pw == NULL)
  {
    fprintf(
      stderr,
      "error: failed to determine uid for user %s\n",
      user
    );
    return EXIT_FAILURE;
  }
  uid = pw->pw_uid;

  gr = getgrnam(group);
  if (gr == NULL)
  {
    fprintf(
      stderr,
      "error: failed to determine gid for group %s\n",
      group
    );
    return EXIT_FAILURE;
  }
  gid = gr->gr_gid;


  fd = inotify_init();
  if (fd < 0)
  {
    perror("inotify_init");
  }

  watchlist_t watchlist =
  {
    .fd = fd,
    .events = EVENTS,
    .size = 0,
    .used = 0,
    .list = NULL
  };

  for (i = optind; i < argc; i++)
  {
    if (scan(argv[i], &watchlist, uid, gid))
    {
      perror("scan");
      return EXIT_FAILURE;
    }
  }

  if (daemonize)
  {
    debug_print("%s\n", "forking");
    pid = fork();
  }
  else
  {
    debug_print("%s\n", "not forking");
    pid = getpid();
  }

  if (pid)
  {
    debug_print("PID is %d\n", pid);

    if (pid_path != NULL)
    {
      debug_print("writing PID (%d) to %s\n", pid, pid_path);
      f = fopen(pid_path, "wb");
      if (f == NULL)
      {
        fprintf(
          stderr,
          "error: failed to open PID file %s\n",
          pid_path
        );
        return EXIT_FAILURE;
      }
      fprintf(f, "%d", pid);
      fclose(f);
    }

    if (daemonize)
    {
      debug_print("%s\n", "in parent, exiting...");
      exit(EXIT_SUCCESS);
    }
  }

  while ((l = read(fd, queue, EVENT_Q_LEN)))
  {
    i = 0;
    while (i < l)
    {
      e = (struct inotify_event *) &queue[i];
      i += EVENT_SIZE + e->len;


      // triggered for items in watched directory: e->name is set
      if (e->mask & (IN_CREATE | IN_MOVED_TO))
      {
        debug_print("event: %s\n", "IN_CREATE | IN_MOVED_TO");
        strcpy(tmppath, get_wd_path(e->wd, &watchlist));
        j = strlen(tmppath);
        strcpy(tmppath + j, e->name);
        scan(tmppath, &watchlist, uid, gid);
      }


      else if (e->mask & IN_ATTRIB)
      {
        debug_print("event: %s\n", "IN_ATTRIB");
        strcpy(tmppath, get_wd_path(e->wd, &watchlist));
        if (e->name != NULL)
        {
          j = strlen(tmppath);
          strcpy(tmppath + j, e->name);
        }
        adjust_attrib(tmppath, uid, gid);
      }


      else if (e->mask & IN_DELETE_SELF)
      {
        debug_print("event: %s\n", "IN_DELETE_SELF");
        remove_wd(e->wd, &watchlist);
      }

      /*
         Remove directories that get moved. No information is provided about
         the new location, which may be outside of the user-specified paths.
         If the directory was moved to another location within the watched
         hierarchy then it will be caught by IN_MOVED_TO above and re-added.
      */
      else if (e->mask & IN_MOVE_SELF)
      {
        debug_print("event: %s\n", "IN_MOVE_SELF");
        unwatch_dir(e->wd, &watchlist);
      }


      // Kill the watchlist and start over if the queue overflows.
      if (e->mask & IN_Q_OVERFLOW)
      {
        debug_print("event: %s\n", "IN_Q_OVERFLOW");
        unwatch_all(&watchlist);
        for (i = optind; i < argc; i++)
        {
          if (scan(argv[i], &watchlist, uid, gid))
          {
            perror("scan");
            return EXIT_FAILURE;
          }
        }
      }
    }
  }
  unwatch_all(&watchlist);
  free(queue);
  return EXIT_SUCCESS;
}
